#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <linux/unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

//#include <linux/dirent.h>
struct dirent {
        long            d_ino;
        off_t           d_off;
        unsigned short  d_reclen;
        char            d_name[256]; /* We must not include limits.h! */
} *d;

_syscall3(int, getdents, uint, fd, struct dirent *, dirp, uint, count);


#define MAXNUM	100

#define RENUMBER	1
#define LIM_SIZE	2
#define LIM_NUMBER	4

char *argv0;		//Name of the program (for error msgs)
char buf[1024];		//For copying, and making temp filenames
char buf2[1024];	//For making temp filenames

int getvernum(char *filename, char *vername) {
	int n=0;

	// Leading dot?
	if (*vername++ != '.') return -1;

	// Matching name?
	while (*filename && *vername) {
		if (*filename++ != *vername++) return -1;
	}

	// Comma?
	if (*vername++ != ',') return -1;

	// Get Number
	do {
		if (*vername<'0' || *vername>'9') return -1;
		n = n*10 + *vername - '0';
	} while (*(++vername));

	return n;
}

void getversions(char *name, int verarray[], int *numver) {
        int fd, size, cur, num, i;
        struct dirent *d;

	// Open directory
	//
	fd = open(".",0);
        if (fd < 0) {
		perror("Can\'t open directory");
		exit(1);
	}

	num=0;

	// Read Directory
	//
        while(1) {
                size = getdents(fd, (struct dirent *)buf, 1024);
                if (size < 0) {
			perror("Can\'t read directory");
			exit(1);
		}
                if (size == 0) break;
                cur = 0;
                while(cur < size) {
                        d = (struct dirent *) (buf+cur);

                        //printf("Name: %s   ",d->d_name);

			// Compare against filename
			i = getvernum(name, d->d_name);
			if (i>=0) {
				if (++num == MAXNUM) {
					printf("Buffer overflow; too many versions\n");
					exit(1);
				}
				verarray[num-1]=i;
			}

                        cur += d->d_reclen;
                }
        }
	*numver=num;
}

int qsort_compar(const void *aa, const void *bb) {
	int a = *((int*)aa);
	int b = *((int*)bb);
	if (a>b) return 1;
	if (a<b) return -1;
	return 0;
}

void purge(char *argv1, int limit, int renum_start, int flags) { 
	int fd, numver, ver[MAXNUM],i,j;
	char *dir;

	// Does file Exist?
	// FIXME: Should I enforce this?
	fd = open(argv1,0);
	if (fd < 0) {
		printf("%s: file \"%s\" not found\n",argv0, argv1);
		exit(1);
	}
	else close(fd);

	dir = strrchr(argv1,'/');
	if (dir == NULL) {}
	else if (dir == argv1) { chdir("/"); argv1=dir+1; }
	else {
		*dir = 0;
		chdir(argv1);
		argv1=dir+1;
	}

	// What are the available versions?
	getversions(argv1, ver, &numver);

	// Sort in ascending order
	qsort((void*)ver, numver, sizeof(int), qsort_compar);

	// Convert size limits into number limits
	if (flags & LIM_SIZE) {
		int sum=0;
		struct stat stbuf;
		assert(!(flags & LIM_NUMBER));
		for (i=numver-1; i>=0; i--) {
			sprintf(buf,".%s,%d",argv1,ver[i]);
			fd = stat(buf, &stbuf);
			if (fd != 0) { perror(argv1); exit(1); }
			sum+=stbuf.st_size;
			if (sum>limit) {
				limit = numver-i-1;
				break;
			}
		}
		if (i == -1) limit=numver;
	}
	else assert(flags & LIM_NUMBER);
	
	if (numver < limit) {
		//printf("%s: Cannot keep %d versions: only %d versions available\n",
		//	argv0,num,numver);
		//exit(1);
		//FIXME: This condidion is not really an error... is it?
		limit = numver;
	}

	// Purge excess versions
	for(i=0; i<numver-limit; i++) {
		sprintf(buf,".%s,%d",argv1,ver[i]);
		//printf("Unlink:%s\n",buf);
		fd = unlink(buf);
		if (fd != 0) perror(buf);
	}

	// Renumber the remaining ones
	if (flags & RENUMBER) {
		for(i=numver-limit, j=renum_start; i<numver; i++,j++) {
			sprintf(buf,".%s,%d",argv1,ver[i]);
			sprintf(buf2,".%s,%d",argv1,j);
			//printf("ren %s %s\n",buf,buf2);
			fd = rename(buf,buf2);
			if (fd != 0) {
				perror(buf);
				exit(1);
			}
		}
	}

	return;
}

int getnum(char* s) {
	int num=0;
	do {
		if (*s<'0' || *s>'9') {
			if (*(s+1) == 0) {
				switch(*s) {
					case 'm':
					case 'M': num *= 1024;

					case 'k':
					case 'K': num *= 1024; return num;

					default: //nothing
				}
			}
			printf("%s: Invalid character \'%c\' in number\n",argv0,*s);
			exit(1);
		}
		num= num*10 + *(s++) - '0';
	} while (*s);
	return num;
}

void usage(void) {
	printf("Usage: %s <command>*\n",argv0);
	printf("       <command>  -->  <filename> | <opt>\n");
	printf("       <opt>      -->  -o | -[rns] <num>\n");
	printf("       <num>      -->  [digit]*[km]\n");
	printf("\n");
	printf("       -r : renumber starting at <num>\n");
	printf("       -o : disable renumbering\n");
	printf("       -n : keep the last <num> versions\n");
	printf("       -s : use atmost <num> bytes of storage for versions\n");
	printf("       Commands are processed from left to right\n");
}

int main(int argc, char **argv) {
	int i, limit=0, renum=0, flags=LIM_NUMBER, num;
	char opt;

	argv0 = argv[0];
	if (argc == 1) {
		usage();
		exit(0);
	}
	for (i=1; i<argc; ) {
		//printf("Arg %d: %s\n",i,argv[i]);
		if (argv[i][0] != '-') purge(argv[i++], limit, renum, flags);
		else {
			// Get Option
			opt = argv[i][1];
			if ((opt == 'o') || (opt == 'O')) {
				flags &= ~RENUMBER;
				i++;
				break;
			}

			// Get the number
			if (argv[i++][2] == 0) {
				if (i == argc) {
					printf("%s: Error on arg %d\n",argv0,i-1);
					usage();
					exit(1);
				}
				num=getnum(argv[i++]);
			}
			else {
				num=getnum(argv[i-1]+2);
			}

			// What option was that?
			switch(opt) {
			case 'r':
			case 'R': renum=num;
				  flags |= RENUMBER;
				  break;

			case 's':
			case 'S': limit=num;
				  flags = ((flags & ~LIM_NUMBER) | LIM_SIZE);
				  break;

			case 'n':
			case 'N': limit=num;
                                  flags = ((flags & ~LIM_SIZE) | LIM_NUMBER); 
                                  break;

			default:  printf("%s: Error on arg %d\n",argv0,i-1);
				  usage();
				  exit(1);
			}
		}
	}
	return 0;
}

