#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <linux/unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

//#include <linux/dirent.h>
struct dirent {
        long            d_ino;
        off_t           d_off;
        unsigned short  d_reclen;
        char            d_name[256]; /* We must not include limits.h! */
} *d;

_syscall3(int, getdents, uint, fd, struct dirent *, dirp, uint, count);

#define BUF_SIZE 4096
#define MAXNUM 100

int getvernum(char *filename, char *vername) {
	int n=0;

	// Leading dot?
	if (*vername++ != '.') return -1;

	// Matching name?
	while (*filename && *vername) {
		if (*filename++ != *vername++) return -1;
	}

	// Comma?
	if (*vername++ != ',') return -1;

	// Get Number
	do {
		if (*vername<'0' || *vername>'9') return -1;
		n = n*10 + *vername - '0';
	} while (*(++vername));

	return n;
}

char buf[1024], newname[256];
char cpybuf[BUF_SIZE];

void getversions(char *name, int verarray[], int *numver) {
        int fd, size, cur, num, i;
        struct dirent *d;

	// Open directory
	//
	fd = open(".",0);
        if (fd < 0) {
		perror("Can\'t open directory");
		exit(1);
	}

	num=0;

	// Read Directory
	//
        while(1) {
                size = getdents(fd, (struct dirent *)buf, 1024);
                if (size < 0) {
			perror("Can\'t read directory");
			exit(1);
		}
                if (size == 0) break;
                cur = 0;
                while(cur < size) {
                        d = (struct dirent *) (buf+cur);

                        //printf("Name: %s   ",d->d_name);

			// Compare against filename
			i = getvernum(name, d->d_name);
			if (i>=0) {
				if (++num == MAXNUM) {
					printf("Buffer overflow; too many versions\n");
					exit(1);
				}
				verarray[num-1]=i;
			}

                        cur += d->d_reclen;
                }
        }
	*numver=num;
}

int qsort_compar(const void *aa, const void *bb) {
	int a = *((int*)aa);
	int b = *((int*)bb);
	if (a>b) return -1;
	if (a<b) return 1;
	return 0;
}

int main(int argc, char **argv) {
	int fd, src, num, numver, ver[MAXNUM], stat, written, vflag = 0, rel = 0;
	char *filename, *fpos;

	// Check Arguments
	if (argc>4 || argc<2) {
		printf("Usage: %s <filename> [-o] [steps]\n",argv[0]);
		printf("-o: <filename> will not be versioned\n");
		exit(1);
	}
	if(argc > 2) {
		int i;

		for(i = 2; i < argc; i++) {
			if((argv[i][0] == '-') && !isdigit(argv[i][1])) {
				if(argv[i][1] != 'o') {
					printf("%s: Option %s not recognized.\n", argv[0], argv[3]);
					exit(1);
				}
				vflag = 1;
			}
			else {
				int j, start = 0;

				if(argv[i][0] == '-') start = 1;

				for(j = start; j < strlen(argv[i]); j++) if(!isdigit(argv[i][j])) {
					printf("%s: Invalid number of steps: %s\n", argv[0], argv[i]);
					exit(1);
				}
				num = atoi(argv[i]);
				if(num < 0) {
					rel = 1;
					num = -num;
				}
				else if(start && !num) {
					printf("%s: Invalid number of steps: %s\n", argv[0], argv[i]);
					exit(1);
				}
			}
		}
	}
		
	// Does file Exist?
	fd = open(argv[1],0);
	if (fd < 0) {
		printf("%s: file \"%s\" not found\n",argv[0], argv[1]);
		exit(1);
	}
	else close(fd);

	filename = strrchr(argv[1],'/');
	if (filename == NULL) {}
	else if (filename == argv[1]) { chdir("/"); argv[1]=filename+1; }
	else {
		*filename = 0;
		chdir(argv[1]);
		argv[1]=filename+1;
	}

	// What are the available versions?
	getversions(argv[1], ver, &numver);
	if (rel && (numver < num)) {
		printf("%s: Cannot revert %d steps: only %d versions available\n",
			argv[0],num,numver);
		exit(1);
	}

	// Sort the version numbers
	qsort((void*)ver, numver, sizeof(int), qsort_compar);

	if(!rel) {
		int i;

		for(i = 0; i < numver; i++) {
			if(ver[i] == num) break;
			if((ver[i] < num) || (i == numver - 1)) {
				printf("%s: Version number %d does not exist for file %s\n", argv[0], num, argv[1]);
				exit(1);
			}
		}
	}
	
	if(vflag) {
		if(tmpnam(buf) == NULL) {
			perror("tmpnam");
			exit(1);
		}
		stat = rename(argv[1], buf);
		if(stat != 0) {
			perror("rename");
			exit(1);
		}
		unlink(buf);
	}

	strcpy(newname, "temp.XXXXXX");
	fd = mkstemp(newname);

	if(fd < 0) {
		perror("mktemp");
		exit(1);
	}
	if(rel) sprintf(buf, ".%s,%d", argv[1], ver[num - 1]);
	else sprintf(buf, ".%s,%d", argv[1], num);
	src = open(buf, O_RDONLY);
	if(src < 0) {
		perror(buf);
		exit(1);
	}
		     
	do {
		stat = read(src, (void *)cpybuf, BUF_SIZE);
		if (stat < 0) {
			perror("copy/read");
			exit(1);
		}
		if(stat == 0) break;
		
		fpos = cpybuf;
		while (1) {
			written = write(fd, (void *)fpos, stat);
			if(written < 0) {
				perror("copy/write");
				exit(1);
			}
			if((stat -= written) == 0) break;
			fpos += written;
		}
	} while (1);
	
	if((stat = close(src))) {
		perror("close");
		exit(1);
	}
	if((stat = close(fd))) {
		perror("close");
		exit(1);
	}
	
	fd = rename(newname,argv[1]);
	if (fd != 0) perror("rename");
	
	return fd;
}
