#ifndef _VL_H_
#define _VL_H_

/*****************************************************************************/
/* Include files                                                             */
/*****************************************************************************/
#include <linux/kernel.h>
#include <sys/syscall.h>
#include <linux/sched.h>
#include <linux/file.h>
#include <linux/malloc.h>
#include <asm/uaccess.h>
#include <linux/smp_lock.h>
#include <asm/semaphore.h>
#include <linux/fs.h>


/*****************************************************************************/
/* Data structure definitions                                                */
/*****************************************************************************/

// Cache entry for per directory policies
//
struct dpolicy {
	unsigned long tag_inode;
	kdev_t tag_device;
	unsigned char numlimit, isize, esize;
	unsigned char onwrite  : 1;
	unsigned char onunlink : 1; 
	unsigned char rdonly   : 1;
	char *incl, *excl;
};

// Cache entry for open file policy information
//
struct fpolicy {
	struct fpolicy *next;
	struct file *tag;
	unsigned char numlimit;
	unsigned char rdonly       : 1;
	unsigned char isdirty      : 1;
	unsigned char ispolicyfile : 1;
};

// Struct returned by get_version_number()
//
struct verinfo {
	int lo;        // min version
	int hi;        // max version
	int num;       // number of versions
};

// Needed for mmap - not defined in any header file
//
struct mmap_arg_struct {
	unsigned long addr;
	unsigned long len;
	unsigned long prot;
	unsigned long flags;
	unsigned long fd;
	unsigned long offset;
};


/*****************************************************************************/
/* Constants                                                                 */
/*****************************************************************************/

// DPOL_TABLE_SIZE must be 2^n.  DPOL_MASK is 2^n - 1 (n consecutive 1's)
//
#define DPOL_TABLE_SIZE  64
#define DPOL_MASK        0x3f
#define POLICY_DATA_SIZE (sizeof(struct dpolicy) - sizeof(unsigned long) - \
			  sizeof(kdev_t) - (2 * sizeof(char *)))
#define DPOL_DATA(dp)    (&(dp->numlimit))


/*****************************************************************************/
/* Function prototypes                                                       */
/*****************************************************************************/

// Defined in open.c
//
int k_open(char *filename, int flags, int mode);
asmlinkage int vl_open(const char *filename, int flags, int mode);
asmlinkage int vl_creat(const char *pathname, int mode);
asmlinkage int vl_close(unsigned int fd);
asmlinkage int vl_unlink(const char *pathname);
asmlinkage int vl_rename(const char *oldname, const char *newname);
asmlinkage int vl_truncate(const char *path, unsigned long length);
asmlinkage int vl_ftruncate(unsigned int fd, unsigned long length);

// Defined in regex.c
// 
int match(const char *filename, char *regex);

// Defined in policy.c
//
void invalidate_dpolicy(struct dentry *dir);
struct dpolicy *get_dpolicy(struct dentry *dir);
int read_dpolicy(struct dentry *dir, struct dpolicy *dpol);
struct fpolicy *ins_fpolicy(struct file *file, struct dpolicy *dpol);
int rem_fpolicy(struct file *file);
struct fpolicy *get_fpolicy(struct file *file);

// Defined in version.c
//
int create_version(unsigned int fd, char *filename);
int get_version_number(struct verinfo*, struct dentry *dentry, const char *name);

// Defined in write.c
//
asmlinkage ssize_t vl_write(unsigned int fd, const char *buf, size_t count);
asmlinkage ssize_t vl_writev(unsigned long fd, const struct iovec * vector,
			     unsigned long count);
asmlinkage int vl_mmap(struct mmap_arg_struct *arg);


// Defined in cleanup.c
//
int cleanup(struct dentry *fdent, unsigned char numlimit, struct verinfo *vinf,
	    char *buf);

/*****************************************************************************/
/* External global variables                                                 */
/*****************************************************************************/
extern struct dpolicy ptab[DPOL_TABLE_SIZE];
extern char *vfile;

/*****************************************************************************/
/* External function pointer variables                                       */
/*****************************************************************************/
extern asmlinkage off_t (*sys_lseek)(unsigned int, off_t, unsigned int);
extern asmlinkage ssize_t (*sys_read)(unsigned int, char *, size_t);

extern asmlinkage int (*old_sys_unlink)(const char *);
extern asmlinkage int (*old_sys_open)(const char *, int, int);
extern asmlinkage int (*old_sys_creat)(const char *, int);
extern asmlinkage ssize_t (*old_sys_write)(unsigned int, const char *, size_t);
extern asmlinkage int (*old_sys_close)(unsigned int);
extern asmlinkage int (*old_sys_rename)(const char *, const char *);
extern asmlinkage int (*old_sys_truncate)(const char *, unsigned long);
extern asmlinkage int (*old_sys_ftruncate)(unsigned int, unsigned long);
extern asmlinkage ssize_t (*old_sys_writev)(unsigned long fd, const struct iovec * vector,
			     unsigned long count);
extern asmlinkage int (*old_sys_mmap)(struct mmap_arg_struct *arg);
     
#endif











