#include "vl.h"
#include <linux/module.h>

//
// FIXME: In order of priority
//
// A I   A=Assigned (A=Adam/P=param/N=not before paper), I = in progress (W = working, D = Done)
// ---------------------------------------------------------------------------------------------
// D     Purge updates for more options, command line flexibility
// A     Relative version numbering
// B     Full error checking for getname()
// A     Empty open file table on exit
// A     Pinning
// A     Hash the open file table
// B     Locking
// ?     Handle version files sanely in other system calls
// ?     
//

// Decalaration of the kernel system call table
//
extern void *sys_call_table[];

// Declaration of miscellaneous system calls that we will extract
// from the system call table
//
asmlinkage off_t (*sys_lseek)(unsigned int, off_t, unsigned int);
asmlinkage ssize_t (*sys_read)(unsigned int, char *, size_t);

// Pointers to old system calls that will be replaced
//
asmlinkage int (*old_sys_open)(const char *, int, int);
asmlinkage int (*old_sys_creat)(const char *, int);
asmlinkage ssize_t (*old_sys_write)(unsigned int, const char *, size_t);
asmlinkage ssize_t (*old_sys_writev)(unsigned long fd, const struct iovec * vector,
				     unsigned long count);
asmlinkage int (*old_sys_close)(unsigned int);
asmlinkage int (*old_sys_unlink)(const char *);
asmlinkage int (*old_sys_rename)(const char *, const char *);
asmlinkage int (*old_sys_truncate)(const char *, unsigned long);
asmlinkage int (*old_sys_ftruncate)(unsigned int, unsigned long);
asmlinkage int (*old_sys_mmap)(struct mmap_arg_struct *arg);

int init_module()
{
	// Get pointers to some system calls that we need to use
	//
	sys_lseek = sys_call_table[__NR_lseek];
	sys_read = sys_call_table[__NR_read];

	// Capture the old system calls
	//
	old_sys_open = sys_call_table[__NR_open];
	old_sys_creat = sys_call_table[__NR_creat];
	old_sys_close = sys_call_table[__NR_close];
	old_sys_write = sys_call_table[__NR_write];
	old_sys_writev = sys_call_table[__NR_writev];
	old_sys_unlink = sys_call_table[__NR_unlink];
	old_sys_rename = sys_call_table[__NR_rename];
	old_sys_truncate = sys_call_table[__NR_truncate];
	old_sys_ftruncate = sys_call_table[__NR_ftruncate];
	old_sys_mmap = sys_call_table[__NR_mmap];


	// Swap in the new system calls
	//
	sys_call_table[__NR_open] = vl_open;
	sys_call_table[__NR_creat] = vl_creat;
	sys_call_table[__NR_write] = vl_write;
	sys_call_table[__NR_writev] = vl_writev;
	sys_call_table[__NR_close] = vl_close;
	sys_call_table[__NR_unlink] = vl_unlink;
	sys_call_table[__NR_rename] = vl_rename;
	sys_call_table[__NR_truncate] = vl_truncate;
	sys_call_table[__NR_ftruncate] = vl_ftruncate;
	sys_call_table[__NR_mmap] = vl_mmap;
	
	return 0;
}


/* Remove the module from the kernel, restoring the original system calls */
void cleanup_module()
{
	// Clean out open file table!
	//

	// Remove system calls
	//
	sys_call_table[__NR_open] = old_sys_open;
	sys_call_table[__NR_creat] = old_sys_creat;
	sys_call_table[__NR_write] = old_sys_write;
	sys_call_table[__NR_writev] = old_sys_writev;
	sys_call_table[__NR_close] = old_sys_close;
	sys_call_table[__NR_unlink] = old_sys_unlink;
	sys_call_table[__NR_rename] = old_sys_rename;
	sys_call_table[__NR_truncate] = old_sys_truncate;
	sys_call_table[__NR_ftruncate] = old_sys_ftruncate;
	sys_call_table[__NR_mmap] = old_sys_mmap;
}

/* Verify that it is safe to unload the module */
int can_unload()
{
	if(sys_call_table[__NR_open] != vl_open) return -1;
	if(sys_call_table[__NR_creat] != vl_creat) return -1;
	if(sys_call_table[__NR_close] != vl_close) return -1;
	if(sys_call_table[__NR_write] != vl_write) return -1;
	if(sys_call_table[__NR_writev] != vl_writev) return -1;
	if(sys_call_table[__NR_unlink] != vl_unlink) return -1;
	if(sys_call_table[__NR_rename] != vl_rename) return -1;
	if(sys_call_table[__NR_truncate] != vl_truncate) return -1;
	if(sys_call_table[__NR_ftruncate] != vl_ftruncate) return -1;
	if(sys_call_table[__NR_mmap] != vl_mmap) return -1;
	return 0;
}
