#include "vl.h"
#define DEBUG

asmlinkage ssize_t vl_write(unsigned int fd, const char * buf, size_t count) {
	int e1, e2;
	char *vername = __getname();

	if ((e1 = create_version(fd, vername)) < 0)
		printk("VL: create_version() failed; Error:%d\n", e1);
	e2 = old_sys_write(fd, buf, count);

	if ((e2 < 0) && (e1 >= 0)) {
#ifdef DEBUG
		printk("    Write Failed; fd=%d, err=%d\n", fd, e2);
#endif
		e1 = old_sys_unlink(vername);
		if(e1 < 0) printk("VL: unlink() following failed write()"
				  " failed; Error %d\n", e1);
	}

	putname(vername);
	return e2;
}



asmlinkage ssize_t vl_writev(unsigned long fd, const struct iovec * vector,
			     unsigned long count) {
	int e1, e2;
	char *vername = __getname();

	if ((e1 = create_version(fd, vername)) < 0)
		printk("VL: create_version() failed; Error:%d\n",e1);
	e2 = old_sys_writev(fd, vector, count);

	if ((e2 < 0) && (e1 >= 0)) {
#ifdef DEBUG
		printk("    Write Failed; fd=%ld, err=%d\n", fd, e2);
#endif
		e1 = old_sys_unlink(vername);
		if(e1 < 0) printk("VL: unlink() following failed writev()"
				  " failed; Error %d\n", e1);
	}

	putname(vername);
	return e2;
}


asmlinkage int vl_mmap(struct mmap_arg_struct *arg) {

	int e1, e2;
	char *vername = __getname();

	// fd == -1 is a POSIX.1b defined feature that has nothing to do with
	// a file (it emulates /dev/zero).  Therefore, we don't need to version it
	//
	if (arg->fd == -1) return old_sys_mmap(arg);

	// Pessimistically create copies on all mmaps
	//
	if ((e1 = create_version(arg->fd, vername)) < 0)
		printk("VL: create_version() failed; Error:%d\n",e1);
	e2 = old_sys_mmap(arg);

	if ((e2 < 0) && (e1 >= 0)) {
#ifdef DEBUG
		printk("    Write Failed; fd=%ld, err=%d\n", arg->fd, e2);
#endif
		e1 = old_sys_unlink(vername);
		if(e1 < 0) printk("VL: unlink() following failed mmap()"
				  " failed; Error %d\n", e1);
	}

	putname(vername);
	return e2;
}








